# -*- coding: utf-8 -*-
import numpy as np
from scipy.stats import t, truncnorm, gamma, invgamma
# from scipy.interpolate import interp1d, interp2d, RegularGridInterpolator
from scipy.optimize import minimize_scalar
from decimal import Decimal as dec

def truncated_normal(mean, std, low, upp):
    """
    Function for probability distribution:
    returns a frozen truncated normal continuous random variable

    Parameters
    ----------
    mean : float, optional
        mean of the sparing factor distribution
    std : float, optional
        standard deviation of the sparing factor distribution
    low : float, optional
        lower bound of distribution
    upp : float, optional
        upper bound of distribution

    Returns
    -------
    scipy.stats._distn_infrastructure.rv_frozen
        distribution function

    """
    normal = truncnorm((low - mean) / std, (upp - mean) / std,
                        loc=mean, scale=std)

    return normal

def student_t(sf_observed, shape_inv, scale_inv):
    """
    Function for probability updating:
    Computes a posterior predictive sparing factor distribution
    for a list of sparing factors (sf_observed) and an inverse-gamma
    conjugate prior distribution given by the shape_inv and scale_inv 
    parameters. Returns a frozen student's t random variable posterior

    Parameters
    ----------
    sf_observed : list/array
        list of observed sparing factors
    shape_inv : float
        shape of inverse-gamma distribution
    scale_inv : float
        scale of inverse-gamma distrinbution

    Returns
    -------
    scipy.stats._distn_infrastructure.rv_continuous_frozen
        distribution function

    """
    n_sf = len(sf_observed)
    variance = np.var(sf_observed)
    mean = np.mean(sf_observed)

    # update the hyperparameters
    shape_prime = shape_inv + n_sf / 2
    # n_sf to correct for normalisation
    scale_prime = scale_inv + n_sf * variance / 2
    # constitue random variable
    student_t = t(df=2 * shape_prime, loc=mean,
        scale=np.sqrt(scale_prime / shape_prime))
    return student_t

def fit_gamma_prior(sf_data):
    """
    fits the shape and scale parameters for a gamma distribution

    Parameters
    ----------
    sf_data : array
        a nxk matrix with n the amount of patients and k the amount
        of sparing factors per patient

    Returns
    -------
    list
        shape and scale hyperparameter
    """
    stds = np.std(sf_data, axis=1)
    shape, _, scale = gamma.fit(stds, floc=0)
    return [shape, scale]

def fit_invgamma_prior(sf_data):
    """
    fits the shape_inv and scale_inv parameters for an
    inverse-gamma distribution

    Parameters
    ----------
    sf_data : array
        a nxk matrix with n the amount of patients and k the amount
        of sparing factors per patient

    Returns
    -------
    list
        shape_inv and scale_inv hyperparameter
    """
    # use variance instead of standard deviation
    variances = np.var(sf_data, axis=1)
    shape_inv, _, scale_inv = invgamma.fit(variances, floc=0)
    return [shape_inv, scale_inv]

def sf_probdist(X, sf_low, sf_high, sf_stepsize, probability_threshold):
    """
    This function computes a probability distribution
    based on the frozen random variable X

    Parameters
    ----------
    X : scipy.stats._distn_infrastructure.rv_frozen
        random variable
    sf_low : float
        lower bound for sparing factor distribution
    sf_high : float
        upper bound for sparing factor distribution
    step_size : float
        sparing factor step size
    probability_threshold : float
        threshold of the probability distribution

    Returns
    -------
    list
        of sf, and prob
    sf : array
        array with sparing factors
    prob : array
        array with probability to each sparing factor

    """
    # create sample sparing factor array
    sample_sf = np.arange(sf_low, sf_high + sf_stepsize, sf_stepsize)
    n_samples = len(sample_sf)

    # sum probability density from cumulative density function in interval
    # to get probability
    half_interval = sf_stepsize/2
    upp_bound = (half_interval - sf_stepsize*1e-6) * np.ones(n_samples)
    low_bound = half_interval * np.ones(n_samples)
    prob = X.cdf(sample_sf + upp_bound) - X.cdf(sample_sf - low_bound)

     # get rid of all probabilities below given threshold
    probability = prob[prob > probability_threshold]
    sf = sample_sf[prob > probability_threshold]
    return [sf, probability]

def std_posterior(sf_observed, shape, scale):
    """
    Function for probability updating:
    Computes a maximum a priori estimation of the standard deviation
    for a list of sparing factors (sf_observed) and a gamma prior
    distribution given by the shape and scale parameters

    Parameters
    ----------
    sf_observed : list/array
        list/array with k sparing factors
    shape : float
        shape of gamma distribution
    scale : float
        scale of gamma distrinbution

    Returns
    -------
    std_updated : float
        most likely std based on the observed data and gamma prior

    """
    n = len(sf_observed)
    variance_observed = np.var(sf_observed)
    # -------------------------------------------------------------------
    def likelihood(std):
        L = ((std ** (shape - 1)) / (std ** (n - 1)) * np.exp(- std / scale)
            * np.exp(- n * variance_observed / (2 * (std ** 2)))
        )
        return -L
    std_updated = minimize_scalar(likelihood, bounds=(0.001, 0.6), 
                    method='bounded', options={'maxiter':19}).x
    return std_updated

def interpolate(x, x_pred, y_reg):
    """
    calculates y values from interpolated function y(x)

    Parameters
    ----------
    x : array
        x values for interpolated function
    x_fit : array
        x predictor for interpolation
    y_fit : array
        y regressand predictors of interpolation

    Returns
    -------
    y : array
        interpolated values, same shape as x

    """
    y = np.interp(x, x_pred, y_reg)
    return y

def find_exponent(number):
    """
    find exponent of number in order of ten

    Parameters
    ----------
    number : float
        number of interest

    Returns
    -------
    exponent : int
        exponent in order of ten

    """
    return dec(str(number)).normalize().as_tuple().exponent


# -*- coding: utf-8 -*-
"""
this folder holds the adaptfx package
Created on Wed Sep 7 15:28:02 2022
@author: janicweber
"""
from .constants import LOG_BOOL, LOG_LEVEL, LOG_BOOL_LIST, LOG_LEVEL_LIST
from .constants import FULL_DICT, SETTING_DICT, KEY_DICT
from .aft_prompt import *
from .aft_utils import *
from .aft import RL_object
from .maths import *
from .planning import *
from .radiobiology import *
from .visualiser import *
from .reinforce import *
from .reinforce_old import *

__all__ = ['bed_calc_matrix',
            'convert_to_physical',
            'multiple',
            'visualiser',
            'min_oar_bed',
            'min_n_frac',
            'timing']





# -*- coding: utf-8 -*-
import argparse
import json
import adaptfx as afx
import sys
nme = __name__

class RL_object():
    """
    Reinforcement Learning class to check instructions
    of calculation, invoke keys and define
    calculation settings from file
    """
    def __init__(self, instruction_filename):
        try: # check if file can be opened
            with open(instruction_filename, 'r') as f:
                read_in = f.read()
            input_dict= json.loads(read_in)
        except TypeError:
            if isinstance(instruction_filename, dict):
                input_dict = instruction_filename
            else:
                afx.aft_error(f'"{instruction_filename}" not a filename or dict', nme)
        except SyntaxError as syntax_err:
            afx.aft_error(f'error in "{instruction_filename}", {syntax_err}', nme)
        except OSError:
            afx.aft_error(f'No such file: "{instruction_filename}"', nme)
        except ValueError as decode_err:
            afx.aft_error(f'decode error in "{instruction_filename}", {decode_err}', nme)
        except:
            afx.aft_error(f'error in "{instruction_filename}", {sys.exc_info()}', nme)

        try: # check if log flag is existent and boolean
            log_bool = input_dict['log']
        except KeyError as log_err:
            log_bool = afx.LOG_BOOL
            afx.aft_warning(f'no {log_err} flag was given, set to {log_bool}', nme)
        else:
            if not log_bool in afx.LOG_BOOL_LIST:
                afx.aft_error('invalid "log" flag was set', nme)

        try: # check if log flag is existent and boolean
            log_level = input_dict['level']
        except KeyError as level_err:
            log_level = afx.LOG_LEVEL
            afx.aft_warning(f'no {level_err} mode was given, set to {log_level}', nme)
        else:
            if not log_level in afx.LOG_LEVEL_LIST:
                afx.aft_error('invalid "debug" flag was set', nme)

        afx.logging_init(instruction_filename, log_bool, log_level)
        afx.aft_message_info('log level:', log_level, nme)
        afx.aft_message_info('log to file:', log_bool, nme)

        try: # check if algorithm key matches known types
            algorithm = input_dict['algorithm']
        except KeyError as algo_err:
            afx.aft_error(f'{algo_err} key missing in: "{instruction_filename}"', nme)
        else:
            if algorithm not in afx.KEY_DICT:
                afx.aft_error(f'unknown "algorithm" type: "{algorithm}"', nme)
            else:
                afx.aft_message_info('algorithm:', algorithm, nme)

        try: # check if keys exists and is a dictionnary
            raw_keys = input_dict['keys']
        except KeyError as key_err:
            afx.aft_error(f'{key_err} is missing in : "{instruction_filename}"', nme)
        else:
            if not isinstance(raw_keys, dict):
                afx.aft_error('"keys" is not a dictionary', nme)
            # load and check keys
            keys = afx.key_reader(afx.KEY_DICT, afx.FULL_DICT, raw_keys, algorithm)
            afx.aft_message_dict('keys', keys, nme, 1)

        try: # check if settings exist and is a dictionnary
            user_settings = input_dict['settings']
        except KeyError as sets_err:
            afx.aft_warning(f'no {sets_err} were given, set to default', nme, 1)
            settings = afx.SETTING_DICT
            afx.aft_message_dict('settings', settings, nme, 1)
        else:
            if not isinstance(user_settings, dict):
                afx.aft_error('"settings" is not a dictionary', nme)
            # load and check settings
            settings = afx.setting_reader(afx.SETTING_DICT, user_settings)
            afx.aft_message_dict('settings', settings, nme, 1)

        self.algorithm = algorithm
        self.log = log_bool
        self.log_level = log_level
        self.keys = afx.DotDict(keys)
        self.settings = afx.DotDict(settings)

    def optimise(self):
        self.output = afx.multiple(self.algorithm, self.keys, self.settings)

    def fraction_counter(self):
        if self.algorithm == 'frac' and self.keys.fraction == 0:
            # in the case for whole treatment calculations
            n_frac = self.keys.number_of_fractions
            afx.aft_message(f'fractions used: ({self.output.fractions_used}/{n_frac})', nme)
            
    def plot(self):
        out = self.output
        sets = self.settings
        if self.settings.plot_policy:
            afx.plot_val(out.policy.sf, out.policy.states, out.policy.val, out.policy.fractions)
        if self.settings.plot_values:
            afx.plot_val(out.value.sf, out.value.states, out.value.val, out.value.fractions)
        if self.settings.plot_remains:
            afx.plot_val(out.remains.sf, out.remains.states, out.remains.val, out.remains.fractions)

        if sets.plot_policy or sets.plot_values or sets.plot_remains:
            afx.show_plot()
        else:
            afx.aft_message('nothing to plot', nme, 1)

def main():
    """
    CLI interface to invoke the RL class
    """
    start = afx.timing()
    parser = argparse.ArgumentParser(
        description='Calculate optimal dose per fraction dependent on algorithm type'
    )
    parser.add_argument(
        '-f',
        '--filename',
        metavar='',
        help='input instruction filename of dictionary',
        type=str
    )
    parser.add_argument(
        '-g',
        '--gui',
        metavar='',
        help='Provide Graphic User Interface for planning - NOT YET AVAILABLE',
        default=False,
        type=bool
    )
    # In case there is no input show help
    args = parser.parse_args(args=None if sys.argv[1:] else ['--help'])
    plan = RL_object(args.filename)

    afx.aft_message('start session...', nme, 1)
    plan.optimise()
    afx.timing(start)

    # show retrospective dose prescribtion
    afx.aft_message_list('physical tumor dose:', plan.output.physical_doses, nme, 1)
    afx.aft_message_list('tumor dose:', plan.output.tumor_doses, nme)
    afx.aft_message_list('oar dose:', plan.output.oar_doses, nme)

    # show accumulated dose
    afx.aft_message_info('accumulated oar dose:', plan.output.oar_sum, nme, 1)
    afx.aft_message_info('accumulated tumor dose:', plan.output.tumor_sum, nme)

    # number of fractions used
    plan.fraction_counter()

    # show plots
    plan.plot()
    afx.aft_message('close session...', nme, 1)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
2D interpolation. In this program the optimal fraction doses are compueted based on a maximal OAR dose while maximizing tumor BED.
single_fraction allows to compute single fraction doses, while whole_plan computes the doses for a whole treatment plan (when all sparing factors are known).
whole_plan_print prints the doses in a well-aranged manner.
"""


import numpy as np
import scipy as sc
from scipy.stats import gamma, truncnorm


def data_fit(data):
    """
    This function fits the alpha and beta value for the prior

    Parameters
    ----------
    data : array
        a nxk matrix with n the amount of patints and k the amount of sparing factors per patient.

    Returns
    -------
    list
        alpha and beta hyperparameter.
    """
    standard_devs = data.std(axis=1)
    alpha, loc, beta = gamma.fit(standard_devs, floc=0)
    return [alpha, beta]


def get_truncated_normal(mean=0, sd=1, low=0.01, upp=10):
    """
    produces a truncated normal distribution

    Parameters
    ----------
    mean : float, optional
        The default is 0.
    sd : float, optional
        The default is 1.
    low : float, optional
        The default is 0.01.
    upp : float, optional
        The default is 10.

    Returns
    -------
    scipy.stats._distn_infrastructure.rv_frozen
        distribution function.

    """
    return truncnorm((low - mean) / sd, (upp - mean) / sd, loc=mean, scale=sd)


def probdist(X):
    """
    This function produces a probability distribution based on the normal distribution X

    Parameters
    ----------
    X : scipy.stats._distn_infrastructure.rv_frozen
        distribution function.

    Returns
    -------
    prob : list
        list with probabilities for each sparing factor.

    """
    prob = np.zeros(170)
    idx = 0
    for i in np.arange(0.01, 1.71, 0.01):
        prob[idx] = X.cdf(i + 0.004999999999999999999) - X.cdf(i - 0.005)
        idx += 1
    return prob


def argfind(searched_list, value):
    """
    This function is used to find the index of certain values.
    searched_list: list/array with values
    value: value that should be inside the list
    return: index of value

    Parameters
    ----------
    searched_list : list/array
        list in which our searched value is.
    value : float
        item inside list.

    Returns
    -------
    index : integer
        index of value inside list.

    """
    index = min(range(len(searched_list)), key=lambda i: abs(searched_list[i] - value))
    return index


def BED_calc0(dose, ab, sparing=1):
    """
    calculates the BED for a specific dose

    Parameters
    ----------
    dose : float
        physical dose to be delivered.
    ab : float
        alpha-beta ratio of tissue.
    sparing : float, optional
        sparing factor. The default is 1 (tumor).

    Returns
    -------
    BED : float
        BED to be delivered based on dose, sparing factor and alpha-beta ratio.

    """
    BED = sparing * dose * (1 + (sparing * dose) / ab)
    return BED


def BED_calc_matrix(sf, ab, actionspace):
    """
    calculates the BED for an array of values

    Parameters
    ----------
    sf : list/array
        list of sparing factors to calculate the correspondent BED.
    ab : float
        alpha-beta ratio of tissue.
    actionspace : list/array
        doses to be delivered.

    Returns
    -------
    BED : List/array
        list of all future BEDs based on the delivered doses and sparing factors.

    """
    BED = np.outer(sf, actionspace) * (
        1 + np.outer(sf, actionspace) / ab
    )  # produces a sparing factors x actions space array
    return BED


def std_calc(measured_data, alpha, beta):
    """
    calculates the most likely standard deviation for a list of k sparing factors and a gamma conjugate prior
    measured_data: list/array with k sparing factors

    Parameters
    ----------
    measured_data : list/array
        list/array with k sparing factors
    alpha : float
        shape of gamma distribution
    beta : float
        scale of gamma distrinbution

    Returns
    -------
    std : float
        most likely std based on the measured data and gamma prior

    """
    n = len(measured_data)
    std_values = np.arange(0.00001, 0.5, 0.00001)
    likelihood_values = np.zeros(len(std_values))
    for index, value in enumerate(std_values):
        likelihood_values[index] = (
            value ** (alpha - 1)
            / value ** (n - 1)
            * np.exp(-1 / beta * value)
            * np.exp(-np.var(measured_data) / (2 * (value**2 / n)))
        )  # here i have to check whether it is good.
    std = std_values[np.argmax(likelihood_values)]
    return std


def distribution_update(sparing_factors, alpha, beta):
    """
    Calculates the probability distributions for all fractions based on a 6 sparing factor list
    Parameters
    ----------
    sparing_factors : array/list
        list/array with 6 sparing factors
    alpha : float
        shape of inverse-gamma distribution
    beta : float
        scale of inverse-gamme distrinbution

    Returns
    -------
    list
        means and stds of all 5 fractions.

    """
    means = np.zeros(len(sparing_factors))
    stds = np.zeros(len(sparing_factors))
    for i in range(len(sparing_factors)):
        means[i] = np.mean(sparing_factors[: (i + 1)])
        stds[i] = std_calc(sparing_factors[: (i + 1)], alpha, beta)
    means = np.delete(means, 0)
    stds = np.delete(
        stds, 0
    )  # we get rid of the first value as it is only the planning value and not used in a fraction
    return [means, stds]


def value_eval(
    fraction,
    number_of_fractions,
    BED,
    sparing_factors,
    alpha,
    beta,
    abt,
    abn,
    bound,
    min_dose=0,
    max_dose=22.3,
    fixed_prob=0,
    fixed_mean=0,
    fixed_std=0,
):
    """
    calculates the optimal dose for the desired fraction.

    Parameters
    ----------
    fraction : integer
        number of actual fraction (1 for first, 2 for second, etc.)
    number_of_fractions : integer
        number of fractions that will be delivered.
    BED : float
        accumulated BED in OAR (from previous fractions) zero in fraction 1
    sparing_factors : list/array
        list of sparing factor distribution.
    alpha : float
        alpha hyperparameter of std prior derived from previous patients.
    beta : float
        beta hyperparameter of std prior derived from previous patients
    abt : float
        alpha-beta ratio of tumor.
    abn : float
        alpha-beta ratio of OAR.
    bound : float
        upper limit of OAR.
    min_dose : float
        minimal physical doses to be delivered in one fraction. The doses are aimed at PTV 95.
    max_dose : float
        maximal physical doses to be delivered in one fraction. The doses are aimed at PTV 95 .
    fixed_prob : int
        this variable is to turn on a fixed probability distribution. If the variable is not used (0), then the probability will be updated. If the variable is turned to 1, the inserted mean and std will be used for a fixed sparing factor distribution
    fixed_mean: float
        mean of the fixed sparing factor normal distribution
    fixed_std: float
        standard deviation of the fixed sparing factor normal distribution

    Returns
    -------
    list
        returns a list with Values, policies, value of the actual fraction, actual policy, BED delivered to the OAR, BED delivered to the tumor, accumulated OAR BED, accumulated tumor BED

    """
    actual_sparing = sparing_factors[-1]
    if fixed_prob != 1:
        mean = np.mean(
            sparing_factors
        )  # extract the mean and std to setup the sparingfactor distribution
        standard_deviation = std_calc(sparing_factors, alpha, beta)
    if fixed_prob == 1:
        mean = fixed_mean
        standard_deviation = fixed_std
    X = get_truncated_normal(mean=mean, sd=standard_deviation, low=0, upp=1.7)
    prob = np.array(probdist(X))
    sf = np.arange(0.01, 1.71, 0.01)
    sf = sf[prob > 0.00001]  # get rid of all probabilities below 10^-5
    prob = prob[prob > 0.00001]

    BEDT = BEDT = np.arange(BED, bound + 1.6, 1)
    Values = np.zeros(
        ((number_of_fractions - fraction), len(BEDT), len(sf))
    )  # 2d values list with first indice being the BED and second being the sf
    if (
        max_dose > 22.3
    ):  # if the chosen maximum dose is too large, it gets reduced. So the algorithm doesn't needlessly check too many actions
        max_dose = 22.3
    if min_dose > max_dose:
        min_dose = max_dose - 0.1
    actionspace = np.arange(min_dose, max_dose + 0.01, 0.1)
    policy = np.zeros(((number_of_fractions - fraction), len(BEDT), len(sf)))
    upperbound = bound + 1

    delivered_doses = BED_calc_matrix(sf, abn, actionspace)
    BEDT_rew = BED_calc_matrix(
        1, abt, actionspace
    )  # this is the reward for the dose deposited inside the tumor.
    BEDT_transformed, meaningless = np.meshgrid(BEDT_rew, np.zeros(len(sf)))

    for index, frac_state in enumerate(
        np.arange(fraction, number_of_fractions + 1)
    ):  # We have number_of fraction fractionations with 2 special cases 0 and number_of_fractions-1 (last first fraction)
        if (
            index == number_of_fractions - 1
        ):  # first state with no prior dose delivered so we dont loop through BEDT
            future_bed = BED + delivered_doses
            future_bed[
                future_bed > bound
            ] = upperbound  # any dose surpassing the upper bound will be set to the upper bound which will be penalized strongly
            value_interpolation = sc.interpolate.interp2d(sf, BEDT, Values[index - 1])
            future_value = np.zeros(len(sf) * len(actionspace) * len(sf)).reshape(
                len(sf), len(actionspace), len(sf)
            )
            for actual_sf in range(0, len(sf)):
                future_value[actual_sf] = value_interpolation(sf, future_bed[actual_sf])
            future_values_prob = (future_value * prob).sum(
                axis=2
            )  # in this array are all future values multiplied with the probability of getting there. shape = sparing factors x actionspace
            penalties = np.zeros(future_bed.shape)
            penalties[
                future_bed > bound
            ] = (
                -1000
            )  # penalizing in each fraction is needed. If not, once the algorithm reached the upper bound, it would just deliver maximum dose over and over again
            Vs = future_values_prob + BEDT_transformed + penalties

            actual_policy = Vs.argmax(axis=1)
            actual_value = Vs.max(axis=1)

        else:
            if (
                index == number_of_fractions - fraction
            ):  # if we are in the actual fraction we do not need to check all possible BED states but only the one we are in
                if fraction != number_of_fractions:
                    future_bed = BED + delivered_doses
                    overdosing = (future_bed - bound).clip(min=0)
                    penalties_overdose = (
                        overdosing * -1000
                    )  # additional penalty when overdosing is needed when choosing a minimum dose to be delivered
                    future_bed[
                        future_bed > bound
                    ] = upperbound  # any dose surpassing the upper bound will be set to the upper bound which will be penalized strongly
                    value_interpolation = sc.interpolate.interp2d(
                        sf, BEDT, Values[index - 1]
                    )
                    future_value = np.zeros(
                        len(sf) * len(actionspace) * len(sf)
                    ).reshape(len(sf), len(actionspace), len(sf))
                    for actual_sf in range(0, len(sf)):
                        future_value[actual_sf] = value_interpolation(
                            sf, future_bed[actual_sf]
                        )
                    future_values_prob = (future_value * prob).sum(
                        axis=2
                    )  # in this array are all future values multiplied with the probability of getting there. shape = sparing factors x actionspace
                    penalties = np.zeros(future_bed.shape)
                    penalties[
                        future_bed > bound
                    ] = (
                        -1000
                    )  # penalizing in each fraction is needed. If not, once the algorithm reached the upper bound, it would just deliver maximum dose over and over again
                    Vs = (
                        future_values_prob
                        + BEDT_transformed
                        + penalties
                        + penalties_overdose
                    )
                    actual_policy = Vs.argmax(axis=1)
                    actual_value = Vs.max(axis=1)
                else:
                    best_action = (
                        -sf + np.sqrt(sf**2 + 4 * sf**2 * (bound - BED) / abn)
                    ) / (2 * sf**2 / abn)
                    if BED > bound:
                        best_action = np.ones(best_action.shape) * min_dose
                    best_action[best_action < min_dose] = min_dose
                    best_action[best_action > max_dose] = max_dose
                    actual_policy = best_action * 10
                    actual_value = BED_calc0(
                        best_action, abt
                    )  # we do not need to penalize, as this value is not relevant.
            else:
                for bed_index, bed_value in enumerate(
                    BEDT
                ):  # this and the next for loop allow us to loop through all states
                    future_bed = delivered_doses + bed_value
                    overdosing = (future_bed - bound).clip(min=0)
                    future_bed[
                        future_bed > bound
                    ] = upperbound  # any dose surpassing 90.1 is set to 90.1
                    if index == 0:  # last state no more further values to add
                        best_action = (
                            -sf
                            + np.sqrt(sf**2 + 4 * sf**2 * (bound - bed_value) / abn)
                        ) / (2 * sf**2 / abn)
                        if bed_value > bound:
                            best_action = np.zeros(best_action.shape)
                        best_action[best_action < min_dose] = min_dose
                        best_action[best_action > max_dose] = max_dose
                        future_bed = BED_calc0(sf, abn, best_action) + bed_value
                        overdosing = (future_bed - bound).clip(min=0)
                        penalties_overdose = (
                            overdosing * -1000
                        )  # additional penalty when overdosing is needed when choosing a minimum dose to be delivered
                        future_bed[
                            future_bed > bound + 0.0001
                        ] = upperbound  # 0.0001 is added due to some rounding problems
                        penalties = np.zeros(future_bed.shape)
                        if bed_value < bound:
                            penalties[future_bed == upperbound] = -1000
                        Values[index][bed_index] = (
                            BED_calc0(best_action, abt) + penalties + penalties_overdose
                        )
                        policy[index][bed_index] = (best_action - min_dose) * 10
                    else:
                        penalties = np.zeros(future_bed.shape)
                        penalties[future_bed == upperbound] = -1000
                        penalties_overdose = (
                            overdosing * -1000
                        )  # additional penalty when overdosing is needed when choosing a minimum dose to be delivered
                        value_interpolation = sc.interpolate.interp2d(
                            sf, BEDT, Values[index - 1]
                        )
                        future_value = np.zeros((len(sf), len(actionspace), len(sf)))
                        for actual_sf in range(0, len(sf)):
                            future_value[actual_sf] = value_interpolation(
                                sf, future_bed[actual_sf]
                            )
                        future_values_prob = (future_value * prob).sum(axis=2)
                        Vs = (
                            future_values_prob
                            + BEDT_transformed
                            + penalties
                            + penalties_overdose
                        )
                        best_action = Vs.argmax(axis=1)
                        valer = Vs.max(axis=1)
                        policy[index][bed_index] = best_action
                        Values[index][bed_index] = valer
    index_sf = argfind(sf, actual_sparing)
    if fraction != number_of_fractions:
        dose_delivered_tumor = BED_calc0(actionspace[actual_policy[index_sf]], abt)
        dose_delivered_OAR = BED_calc0(
            actionspace[actual_policy[index_sf]], abn, actual_sparing
        )
        total_dose_delivered_OAR = dose_delivered_OAR + BED
        actual_dose_delivered = actionspace[actual_policy[index_sf]]
    else:
        dose_delivered_tumor = BED_calc0(actual_policy[index_sf] / 10, abt)
        dose_delivered_OAR = BED_calc0(
            actual_policy[index_sf] / 10, abn, actual_sparing
        )
        total_dose_delivered_OAR = dose_delivered_OAR + BED
        actual_dose_delivered = actual_policy[index_sf] / 10

    return [
        Values,
        policy,
        actual_value,
        actual_policy,
        dose_delivered_OAR,
        dose_delivered_tumor,
        total_dose_delivered_OAR,
        actual_dose_delivered,
    ]


def whole_plan(
    number_of_fractions,
    sparing_factors,
    abt,
    abn,
    alpha,
    beta,
    OAR_limit,
    min_dose=0,
    max_dose=22.3,
    fixed_prob=0,
    fixed_mean=0,
    fixed_std=0,
):
    """
    calculates whole plan given all sparing factors

    Parameters
    ----------
    number_of_fractions : integer
        number of fractions that will be delivered.
    sparing_factors : list/array
        list/array with all observed sparing factors (6 for a 5 fraction plan).
    abt : float
        alpha-beta ratio of tumor.
    abn : float
        alpha-beta ratio of OAR.
    alpha : float
        shape of inverse-gamma distribution
    beta : float
        scale of inverse-gamme distrinbution
    OAR_limit : float
        accumulated BED in OAR (from previous fractions) zero in fraction 1
    min_dose : float
        minimal physical doses to be delivered in one fraction. The doses are aimed at PTV 95.
    max_dose : float
        maximal physical doses to be delivered in one fraction. The doses are aimed at PTV 95.
    fixed_prob : int
        this variable is to turn on a fixed probability distribution. If the variable is not used (0), then the probability will be updated. If the variable is turned to 1, the inserted mean and std will be used for a fixed sparing factor distribution
    fixed_mean: float
        mean of the fixed sparing factor normal distribution
    fixed_std: float
        standard deviation of the fixed sparing factor normal distribution

    Returns
    -------
    List with delivered tumor doses, delivered OAR doses and delivered physical doses.

    """
    total_dose_delivered_OAR = 0
    total_tumor_dose = 0
    tumor_doses = np.zeros(number_of_fractions)
    OAR_doses = np.zeros(number_of_fractions)
    physical_doses = np.zeros(number_of_fractions)
    for looper in range(0, number_of_fractions):
        [
            Values,
            policy,
            actual_value,
            actual_policy,
            dose_delivered_OAR,
            tumor_dose,
            total_dose_delivered_OAR,
            actual_dose_delivered,
        ] = value_eval(
            looper + 1,
            number_of_fractions,
            total_dose_delivered_OAR,
            sparing_factors[: looper + 2],
            alpha,
            beta,
            abt,
            abn,
            OAR_limit,
            min_dose,
            max_dose,
            fixed_prob,
            fixed_mean,
            fixed_std,
        )
        tumor_doses[looper] = tumor_dose
        physical_doses[looper] = actual_dose_delivered
        OAR_doses[looper] = dose_delivered_OAR
        total_tumor_dose += tumor_dose
    return [tumor_doses, OAR_doses, physical_doses]


def whole_plan_print(
    number_of_fractions,
    sparing_factors,
    abt,
    abn,
    alpha,
    beta,
    OAR_limit,
    min_dose=0,
    max_dose=22.3,
    fixed_prob=0,
    fixed_mean=0,
    fixed_std=0,
):
    """
    calculates whole plan given all sparing factors and prints the results

    Parameters
    ----------
    number_of_fractions : integer
        number of fractions that will be delivered.
    sparing_factors : list/array
        list/array with all observed sparing factors (6 for a 5 fraction plan).
    abt : float
        alpha-beta ratio of tumor.
    abn : float
        alpha-beta ratio of OAR.
    alpha : float
        shape of inverse-gamma distribution
    beta : float
        scale of inverse-gamme distrinbution
    OAR_limit : float
        accumulated BED in OAR (from previous fractions) zero in fraction 1
    min_dose : float
        minimal physical doses to be delivered in one fraction. The doses are aimed at PTV 95.
    max_dose : float
        maximal physical doses to be delivered in one fraction. The doses are aimed at PTV 95.
    fixed_prob : int
        this variable is to turn on a fixed probability distribution. If the variable is not used (0), then the probability will be updated. If the variable is turned to 1, the inserted mean and std will be used for a fixed sparing factor distribution
    fixed_mean: float
        mean of the fixed sparing factor normal distribution
    fixed_std: float
        standard deviation of the fixed sparing factor normal distribution

    Returns
    -------
    None.
    """
    [tumor_doses, OAR_doses, physical_doses] = whole_plan(
        number_of_fractions,
        sparing_factors,
        abt,
        abn,
        alpha,
        beta,
        OAR_limit,
        min_dose,
        max_dose,
        fixed_prob,
        fixed_mean,
        fixed_std,
    )
    for i in range(0, number_of_fractions):
        print("Fraction ", (i + 1))
        print("physical dose delivered  = ", physical_doses[i])
        print("tumor dose in BED = ", tumor_doses[i])
        print("OAR dose in BED = ", OAR_doses[i])
    print("total tumor BED = ", np.sum(tumor_doses))
    print("total OAR BED = ", np.sum(OAR_doses))


def single_fraction(
    number_of_fractions,
    sparing_factors,
    accumulated_OAR_BED,
    OAR_limit,
    abt,
    abn,
    alpha,
    beta,
    min_dose=0,
    max_dose=22.3,
    fixed_prob=0,
    fixed_mean=0,
    fixed_std=0,
):
    """
    Parameters
    ----------
    number_of_fractions : integer
        number of fractions that will be delivered.
    sparing_factors : list/array
        list/array with all observed sparing factors (6 for a 5 fraction plan).
    accumulated_OAR_BED : float
        Total BE delivered to the OAR so far.
    OAR_limit : float
        accumulated BED in OAR (from previous fractions) zero in fraction 1
    abt : float
        alpha-beta ratio of tumor.
    abn : float
        alpha-beta ratio of OAR.
    alpha : float
        shape of inverse-gamma distribution
    beta : float
        scale of inverse-gamme distrinbution
    min_dose : float
        minimal physical doses to be delivered in one fraction. The doses are aimed at PTV 95.
    max_dose : float
        maximal physical doses to be delivered in one fraction. The doses are aimed at PTV 95.
    fixed_prob : int
        this variable is to turn on a fixed probability distribution. If the variable is not used (0), then the probability will be updated. If the variable is turned to 1, the inserted mean and std will be used for a fixed sparing factor distribution
    fixed_mean: float
        mean of the fixed sparing factor normal distribution
    fixed_std: float
        standard deviation of the fixed sparing factor normal distribution

    Returns
    -------
    None.

    """
    [
        Values,
        policy,
        actual_value,
        actual_policy,
        dose_delivered_OAR,
        tumor_dose,
        total_dose_delivered_OAR,
        actual_dose_delivered,
    ] = value_eval(
        len(sparing_factors) - 1,
        number_of_fractions,
        accumulated_OAR_BED,
        sparing_factors,
        alpha,
        beta,
        abt,
        abn,
        OAR_limit,
        min_dose,
        max_dose,
        fixed_prob,
        fixed_mean,
        fixed_std,
    )
    print("fraction", (len(sparing_factors) - 1))
    print("physical dose delivered  = ", actual_dose_delivered)
    print("tumor dose in BED = ", tumor_dose)
    print("OAR dose in BED = ", dose_delivered_OAR)



# -*- coding: utf-8 -*-
import numpy as np
from scipy.stats import gamma

def max_action(bed, actionspace, goal, abt=10):
    """
    Computes the maximal dose that can be delivered to the tumor
    in each fraction depending on the actual accumulated dose

    Parameters
    ----------
    bed : float
        accumulated tumor dose so far.
    actionspace : list/array
        array with all discrete dose steps.
    goal : float
        prescribed tumor dose.
    abt : float, optional
        alpha beta ratio of tumor. The default is 10.

    Returns
    -------
    sizer : integer
        gives the size of the resized actionspace to reach
        the prescribed tumor dose.

    """
    bed_actionspace = bed_calc0(actionspace, abt)
    max_action = min(max(bed_actionspace), goal - bed)
    sizer = np.argmin(np.abs(bed_actionspace - max_action))
    if bed_actionspace[sizer] < max_action:
        # make sure that with prescribing max_action
        # reached dose is not below goal
        sizer += 1

    return sizer

def argfind(bedt, value):
    """
    This function is used to find the index of certain values

    Parameters
    ----------
    bedt : list/array
        list of tumor BED in which value is searched.
    value : float
        item inside list.

    Returns
    -------
    index : integer
        index of value inside list.

    """
    index = min(range(len(bedt)), key=lambda i: abs(bedt[i] - value))
    return index

def distribution_update(sparing_factors, alpha, beta):
    """
    Calculates the probability distributions for all
    fractions based on a 6 sparing factor list

    Parameters
    ----------
    sparing_factors : array/list
        list/array with 6 sparing factors
    alpha : float
        shape of inverse-gamma distribution
    beta : float
        scale of inverse-gamme distrinbution

    Returns
    -------
    list
        means and stds of all 5 fractions

    """
    means = np.zeros(len(sparing_factors))
    stds = np.zeros(len(sparing_factors))
    for i in range(len(sparing_factors)):
        means[i] = np.mean(sparing_factors[: (i + 1)])
        stds[i] = std_calc(sparing_factors[: (i + 1)], alpha, beta)
    means = np.delete(means, 0)
    stds = np.delete(
        stds, 0
    )   
    # we get rid of the first value as it is only the planning 
    # value and not used in a fraction
    return [means, stds]

def data_fit(data):
    """
    This function fits the alpha and beta value for the prior

    Parameters
    ----------
    data : array
        a nxk matrix with n the amount of patients and k the amount
        of sparing factors per patient

    Returns
    -------
    list
        alpha and beta hyperparameter
    """
    standard_devs = data.std(axis=1)
    alpha, _, beta = gamma.fit(standard_devs, floc=0)
    return [alpha, beta]


"""
This script opens a GUI to check the resulting gamma distribution. With the
slide-bars one can directly adapt the distribution and check how the shape of
the distribution changes.
"""

import tkinter as tk

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from scipy.stats import gamma

if __name__ == "__main__":

    window = tk.Tk()
    window.title("gamma distribution preview")
    frm_scale = tk.Frame(relief=tk.SUNKEN, borderwidth=3)
    frm_scale.pack()
    frm_plot = tk.Frame()
    frm_plot.pack()
    fig = plt.Figure(figsize=(10, 10), dpi=100)
    canvas = FigureCanvasTkAgg(fig, master=frm_plot)
    canvas.draw
    canvas.get_tk_widget().pack()
    ax = fig.add_subplot(111)
    ax.set_xlabel("standard deviation")
    ax.set_ylabel("probability density")
    ax.set_title("gamma distribution")

    def gamma_plot(varer):
        x_range = np.arange(0.001, 0.4, 0.00001)
        alpha = var_alpha.get()
        beta = var_beta.get()
        ax.clear()
        ax.set_xlabel("standard deviation")
        ax.set_ylabel("probability density")
        ax.set_title("gamma distribution")
        ax.plot(np.sqrt(x_range), gamma.pdf(x_range, alpha, loc=0, scale=beta))
        canvas.draw()

    var_alpha = tk.DoubleVar()
    scl_alpha = tk.Scale(
        frm_scale,
        label="alpha value",
        from_=0.001,
        to=10.00,
        orient=tk.HORIZONTAL,
        length=400,
        showvalue=1,
        tickinterval=1,
        resolution=0.1,
        variable=var_alpha,
        command=gamma_plot,
    )
    scl_alpha.grid(row=0, column=2)
    var_beta = tk.DoubleVar()
    scl_beta = tk.Scale(
        master=frm_scale,
        label="beta value",
        from_=0.000001,
        to=0.0500,
        variable=var_beta,
        orient=tk.HORIZONTAL,
        length=400,
        tickinterval=0.00999,
        showvalue=1,
        resolution=0.001,
        command=gamma_plot,
    )
    scl_beta.grid(row=1, column=2)


window.mainloop()
